import base64
import copy
import datetime
import importlib
import logging
import random
import re
from importlib import util as importlib_util
from pathlib import Path
from typing import Any, Optional, Union

import tomlkit
from tomlkit.toml_document import TOMLDocument

DEFAULT_CONFIG_FILEPATH = Path(__file__).parent.resolve() / "hyrax_default_config.toml"
DEFAULT_USER_CONFIG_FILEPATH = Path.cwd() / "hyrax_config.toml"
# There are only a couple of configuration keys where we would expect to find an
# external library string, so we specify those here.
KEYS_WITH_EXTERNAL_LIBS = ["name", "dataset_class"]

logger = logging.getLogger(__name__)


def config_help(config: TOMLDocument, *args):
    """
    A simple config help function. It's a bit difficult to parse through
    the Tomlkit Table to print just one item such that it would include the comments
    preceding it.

    For now, we support the following cases, and generally print out the entire
    table for the given key.

    Cases:
    - if no args, prints the whole config.
    - if args[0] is a table name, print the whole table
    - if args[0] is not a table, assume it's a key and search
    -- print each one of the tables that it is found in.

    Parameters
    ----------
    config : TOMLDocument
        A configuration dictionary that will be used to search for specified tables
        and keys.

    args : str
        A variable number of string arguments that specify the table name or key
        to search for in the configuration dictionary.
    """

    # Get the config as a dictionary
    config_dict = config.value

    # No tables provided, print the whole config
    if not args:
        print(config.as_string())

    # Table name provided as args[0], print that config table
    if len(args) == 1 and args[0] in config_dict:
        print(f"[{args[0]}]")
        print(config[args[0]].as_string())

    # One arg provided, but it's not a table name.
    # Assume it's a config key and search the config for it.
    # Print each table that it is found in.
    if len(args) == 1 and args[0] not in config_dict:
        matching = find_keys(config_dict, args[0])
        if len(matching):
            tables = [m.split(".")[0] for m in matching]
            print(f"Found '{args[0]}' in the following config sections: [{'], ['.join(tables)}]")
            for t in tables:
                config_help(config, t)
        else:
            print(f"Could not find '{args[0]}' in the config")

    if len(args) == 2:
        if args[0] in config_dict and args[1] in config_dict[args[0]]:
            print(f"[{args[0]}]")
            print(config[args[0]].as_string())
        else:
            print(f"Cannot find ['{args[0]}']['{args[1]}'] in the current configuration.")

    if len(args) > 2:
        print("Too many arguments provided. Expecting 0, 1, or 2 arguments.")
        print("Usage: config.help(['table_name'|'key_name']), config.help('table_name', 'key_name')")


def parse_dotted_key(key: str) -> list[str]:
    """
    Parse a dotted key string, respecting quoted sections.

    Quoted sections (using single or double quotes) are treated as a single key
    component, even if they contain dots. This allows for keys like 'torch.optim.Adam'
    to be used as a single table name in TOML configuration files.

    Parameters
    ----------
    key : str
        The dotted key to parse. Examples:
        - "model.name" -> ['model', 'name']
        - "'torch.optim.Adam'.lr" -> ['torch.optim.Adam', 'lr']
        - '"torch.optim.Adam".lr' -> ['torch.optim.Adam', 'lr']
        - "optimizer.'torch.optim.Adam'.lr" -> ['optimizer', 'torch.optim.Adam', 'lr']

    Returns
    -------
    list[str]
        A list of key components
    """
    pattern = r"""(['"])(.*?)\1|([^.'"]+)"""
    matches = re.findall(pattern, key)
    parts = []

    # For the example input key = '"torch.optim.Adam".lr'
    # `matches` = [('"', 'torch.optim.Adam', ''), ('', '', 'lr')]
    for _, quoted, unquoted in matches:
        if quoted:
            parts.append(quoted)
        elif unquoted:
            parts.append(unquoted)
    return parts


def find_keys(config: dict[str, Any], key_name: str):
    """
    Recursively find all keys in a nested dictionary that match the given key name.

    Parameters
    ----------
    config : dict
        The nested dictionary to search.
    key_name : str
        The name of the key to find.

    Returns
    -------
    list
        A list of matching keys.
    """
    matching_keys = []

    def _find_keys(d, parent_key=""):
        if isinstance(d, dict):
            for k, v in d.items():
                if k == key_name:
                    matching_keys.append((parent_key + "." + k).strip("."))
                _find_keys(v, parent_key + "." + k)

    _find_keys(config)
    return matching_keys


TOMLDocument.help = config_help  # type: ignore


class ConfigManager:
    """A class to manage the runtime configuration for a Hyrax object. This class
    will contain all the logic and methods for reading, merging, and validating
    the runtime configuration.
    """

    # True when we are called from a test, so we maintain isolation from any
    # user config that may be in the cwd of the running test process
    _called_from_test = False

    """
    Hardcoded set of config keys which we know to contain paths, and we resolve to global paths
    during initialization in ConfigManager._resolve_config_paths().
    """
    PATH_CONFIG_KEYS = [
        # TODO: external library config defaults
        # However we define config defaults from external libraries ought allow them to designate config keys
        # which contain relative paths. ultimately these should end up on the list and be resolved.
        ["data_set", "filter_catalog"],
        ["general", "data_dir"],
    ]

    def __init__(
        self,
        runtime_config_filepath: Optional[Union[Path, str]] = None,
        default_config_filepath: Union[Path, str] = DEFAULT_CONFIG_FILEPATH,
    ):
        self.hyrax_default_config: TOMLDocument = ConfigManager.read_runtime_config(default_config_filepath)

        self.runtime_config_filepath = ConfigManager.resolve_runtime_config(runtime_config_filepath)
        if self.runtime_config_filepath is DEFAULT_CONFIG_FILEPATH:
            self.user_specific_config = TOMLDocument()
        else:
            self.user_specific_config = ConfigManager.read_runtime_config(self.runtime_config_filepath)

        self.config = self._render_config(self.user_specific_config, self.hyrax_default_config)
        self.original_config = copy.deepcopy(self.config)

    @staticmethod
    def _render_config(
        user_specific_config: TOMLDocument = None,
        hyrax_default_config: TOMLDocument = None,
    ):
        user_specific_config = user_specific_config if user_specific_config is not None else TOMLDocument()
        hyrax_default_config = hyrax_default_config if hyrax_default_config is not None else TOMLDocument()

        external_library_config_paths = set()
        external_library_config_paths |= ConfigManager._find_external_library_default_config_paths(
            user_specific_config
        )

        # 1) merge all the external library config dictionaries together
        external_default_config = ConfigManager.merge_external_default_configs(external_library_config_paths)

        # 2) merge the external library configs on top of the hyrax defaults
        overall_default_config = ConfigManager.merge_default_configs(
            hyrax_default_config, external_default_config
        )

        # 3) merge the user config on top of the overall defaults
        config = ConfigManager.merge_configs(overall_default_config, user_specific_config)

        ConfigManager._resolve_config_paths(config)
        if not config["general"]["dev_mode"]:
            ConfigManager._validate_runtime_config(config, overall_default_config)

        return config

    def set_config(self, key: str, value: Any):
        """Set a config value at runtime. This modifies the in-memory config object.
        Once the configuration is updated, the entire config is re-rendered to
        ensure that any requested external library default configs are incorporated.

        Parameters
        ----------
        key : str
            The dotted key to set, e.g. "model.name" or "'torch.optim.Adam'.lr"
            Quoted sections (using single or double quotes) are treated as single
            key components, allowing for table names like 'torch.optim.Adam'.

        value : Any
            The value to set the key to.
        """
        keys = parse_dotted_key(key)
        d = self.config
        for k in keys[:-1]:
            d = d[k]
        d[keys[-1]] = value

        self.config = self._render_config(self.config, self.original_config)
        self.original_config = copy.deepcopy(self.config)

    @staticmethod
    def read_runtime_config(config_filepath: Union[Path, str] = DEFAULT_CONFIG_FILEPATH) -> TOMLDocument:
        """Read a single toml file and return a TOMLDocument

        Parameters
        ----------
        config_filepath : Union[Path, str], optional
            The path to the config file, by default DEFAULT_CONFIG_FILEPATH

        Returns
        -------
        TOMLDocument
            The contents of the toml file as a tomlkit.TOMLDocument
        """
        config_filepath = Path(config_filepath)
        parsed_dict: TOMLDocument = TOMLDocument()
        if config_filepath.exists():
            with open(config_filepath, "r") as f:
                parsed_dict = tomlkit.load(f)

        return parsed_dict

    @staticmethod
    def _find_external_library_default_config_paths(runtime_config: dict) -> set:
        """Search for external libraries in the runtime configuration and gather the
        libpath specifications so that we can load the default configs for the libraries.

        Parameters
        ----------
        runtime_config : dict
            The runtime configuration as a tomlkit.TOMLDocument.
        Returns
        -------
        set
            A tuple containing the default configuration Paths for the external
            libraries that are requested in the users configuration file.
        """

        default_config_paths = set()
        for key, value in runtime_config.items():
            if isinstance(value, dict):
                default_config_paths |= ConfigManager._find_external_library_default_config_paths(value)
            else:
                # We expect that values we are interested in will be of type string.
                if key in KEYS_WITH_EXTERNAL_LIBS and isinstance(value, str) and "." in value:
                    external_library = value.split(".")[0]
                    if importlib_util.find_spec(external_library) is not None:
                        try:
                            lib = importlib.import_module(external_library)
                            if lib.__file__ is None:
                                raise RuntimeError()
                            lib_default_config_path = Path(lib.__file__).parent / "default_config.toml"
                            if lib_default_config_path.exists():
                                default_config_paths.add(lib_default_config_path)
                            else:
                                logger.warning(f"Cannot find default_config.toml for {value}.")
                        except ModuleNotFoundError:
                            logger.error(
                                f"External library {lib} not found. Please install it before running."
                            )
                            raise
                    else:
                        raise ModuleNotFoundError(
                            f"External library {external_library} not found. Check installation."
                        )

        return default_config_paths

    @staticmethod
    def merge_external_default_configs(external_default_config_paths):
        """Merge the default configurations from external libraries into the overall
        default configuration.

        Parameters
        ----------
        external_default_config_paths : set
            A set containing the default configuration Paths for the external
            libraries that are requested in the users configuration file.

        Returns
        -------
        dict
            The merged overall default configuration including the external library defaults.
        """
        overall_default_config = TOMLDocument()
        # Merge all external library default configurations first
        for path in external_default_config_paths:
            logger.info(f"Merging external default config from {path}")
            external_library_config = ConfigManager.read_runtime_config(path)
            overall_default_config = ConfigManager.merge_configs(
                overall_default_config, external_library_config
            )

        return overall_default_config

    @staticmethod
    def merge_default_configs(hyrax_defaults, external_defaults):
        """Merge the default configurations of external libraries on top of the
        Hyrax default configuration.

        Parameters
        ----------
        hyrax_defaults : dict
            The default configuration from hyrax.
        external_defaults : dict
            The default configuration from external libraries.

        Returns
        -------
        dict
            The merged overall default configuration including the external library defaults.
        """
        return ConfigManager.merge_configs(hyrax_defaults, external_defaults)

    @staticmethod
    def merge_configs(base_config: dict, overriding_config: dict) -> dict:
        """Merge two config dictionaries with the overriding_config values overriding
        the base_config values.

        Parameters
        ----------
        base_config : dict
            The base configuration with keys that may be overridden by the
            overriding_config.
        overriding_config : dict
            The new configuration values that will override the values in base_config.

        Returns
        -------
        dict
            The merged configuration.
        """

        final_config = base_config.copy()
        for k, v in overriding_config.items():
            if k in final_config and isinstance(final_config[k], dict) and isinstance(v, dict):
                final_config[k] = ConfigManager.merge_configs(base_config[k], v)
            else:
                final_config[k] = v

        return final_config

    @staticmethod
    def _validate_runtime_config(runtime_config: dict, default_config: dict):
        """Recursive helper to check that all keys in runtime_config have a default
        in the merged default_config.

        The two arguments passed in must represent the same nesting level of the
        runtime config and all default config parameters respectively.

        Parameters
        ----------
        runtime_config : dict
            Nested config dictionary representing the runtime config.
        default_config : dict
            Nested config dictionary representing the defaults

        Raises
        ------
        RuntimeError
            Raised if any config that exists in the runtime config does not have a default defined in
            default_config
        """
        for key in runtime_config:
            if key not in default_config:
                msg = f"Runtime config contains key or section '{key}' which has no default defined. "
                msg += f"All configuration keys and sections must be defined in {DEFAULT_CONFIG_FILEPATH}"
                logger.warning(msg)
                continue

            if isinstance(runtime_config[key], dict):
                if not isinstance(default_config[key], dict):
                    msg = f"Runtime config contains a section named '{key}' which is the name of a "
                    msg += "value in the default config. Please choose another name for this section."
                    logger.warning(msg)
                    continue
                ConfigManager._validate_runtime_config(runtime_config[key], default_config[key])

    @staticmethod
    def _resolve_config_paths(runtime_config: dict) -> None:
        """Convert all paths in a runtime config to global paths in the current environment.
        Uses the hardcoded list of paths in ConfigManager.PATH_CONFIG_KEYS

        This mutates the config dictionary passed.

        Parameters
        ----------
        runtime_config : dict
           Current runtime config nested dictionary
        """
        for key_spec in ConfigManager.PATH_CONFIG_KEYS:
            # Recursively look up a list of keys.
            current_dict = runtime_config
            current_key = None
            current_val = None
            # At the end of each loop current_* are always the dict, key, and value of the
            # last lookup.
            for key in key_spec:
                current_key = key
                try:
                    current_val = current_dict[key]
                except KeyError:
                    break

                if isinstance(current_val, dict):
                    current_dict = current_val

            # On the non-break end of the loop we do path resolution, preserving falsy values
            # as false.
            else:
                new_val = str(Path(current_val).expanduser().resolve()) if current_val else False
                current_dict[current_key] = new_val

    @staticmethod
    def resolve_runtime_config(runtime_config_filepath: Union[Path, str, None] = None) -> Path:
        """Resolve a user-supplied runtime config to where we will actually pull config from.

        #. If a runtime config file is specified, we will use that file.
        #. If no file is specified and there is a file named "hyrax_config.toml" in the cwd we will use it.
        #. If no file is specified and there is no file named "hyrax_config.toml" in the cwd we will
           exclusively work off the configuration defaults in the packaged "hyrax_default_config.toml" file.

        Parameters
        ----------
        runtime_config_filepath : Union[Path, str, None], optional
            Location of the supplied config file, by default None

        Returns
        -------
        Path
            Path to the configuration file ultimately used for config resolution. When we fall back to the
            package supplied default config file, the Path to that file is returned.

        Raises
        ------
        FileNotFoundError
            If a runtime config file is specified but does not exist.
        """
        if isinstance(runtime_config_filepath, str):
            runtime_config_filepath = Path(runtime_config_filepath)

        # If a runtime config file is explicitly specified, validate it exists
        if isinstance(runtime_config_filepath, Path) and not runtime_config_filepath.exists():
            raise FileNotFoundError(f"Cannot find config file {runtime_config_filepath}")

        # If a named config exists in cwd, and no config specified on cmdline, use cwd.
        if (
            runtime_config_filepath is None
            and DEFAULT_USER_CONFIG_FILEPATH.exists()
            and not ConfigManager._called_from_test
        ):
            runtime_config_filepath = DEFAULT_USER_CONFIG_FILEPATH

        if runtime_config_filepath is None:
            runtime_config_filepath = DEFAULT_CONFIG_FILEPATH

        return runtime_config_filepath


def create_results_dir(config: dict, postfix: str) -> Path:
    """Creates a results directory for this run.

    Postfix is the verb name of the run e.g. (infer, train, etc)

    The directory is created within the results dir (set with config results_dir)
    and follows the pattern <timestamp>-<postfix>

    The resulting directory is returned.

    Parameters
    ----------
    config : dict
        The full runtime configuration for this run
    postfix : str
        The verb name of the run.

    Returns
    -------
    Path
        The path created by this function
    """
    results_root = Path(config["general"]["results_dir"]).expanduser().resolve()
    # This date format is chosen specifically to create a lexical search order
    # which matches the date order.
    timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    # Generate 4 random ascii characters to avoid collisions from multiple hyrax processes
    # started in a shared filesystem environment.
    random_str = base64.urlsafe_b64encode(random.randbytes(3)).decode("ascii")
    directory = results_root / f"{timestamp}-{postfix}-{random_str}"
    directory.mkdir(parents=True, exist_ok=False)
    return directory


def find_most_recent_results_dir(config: dict, verb: str) -> Optional[Path]:
    """Find the most recent results directory corresponding to a particular verb
    This is a best effort search in the currently configured results root.

    If result directories are created within 1 second of one another this function
    will return one of the directories but it is undefined which one it will return.

    This function may return None indicating it could not find a directory matching
    the query verb
    """
    results_root = Path(config["general"]["results_dir"]).expanduser().resolve()

    max_timestamp = 0
    best_path = None

    for path in results_root.glob(f"*-{verb}-*"):
        if path.is_dir():
            regex = r"([0-9]{8})-([0-9]{6})-.*"
            m = re.match(regex, path.name)

            if m is None:
                continue

            timestamp = int(m[1] + m[2])
            if timestamp > max_timestamp:
                max_timestamp = timestamp
                best_path = path

    return best_path


def log_runtime_config(runtime_config: dict, output_path: Path, file_name: str = "runtime_config.toml"):
    """Log a runtime configuration.

    Parameters
    ----------
    runtime_config : dict
        A dictionary object containing runtime configuration values.
    output_path : str
        The path to put the config file
    file_name : str, Optional
        Optional name for the config file, defaults to "runtime_config.toml"
    """
    with open(output_path / file_name, "w") as f:
        f.write(tomlkit.dumps(runtime_config))



def test_train(loopback_hyrax):
    """
    Simple test that training succeeds with the loopback
    model in use.
    """
    h, _ = loopback_hyrax
    h.train()


def test_train_percent_split(tmp_path):
    """
    Ensure backward compatibility with percent-based splits when the
    configuration provides only a `train` and `infer` model_inputs section
    (no explicit `validate` table). This should exercise the code path
    that creates train/validate splits from a single dataset location.
    """
    import hyrax

    h = hyrax.Hyrax()
    h.config["model"]["name"] = "HyraxLoopback"
    h.config["train"]["epochs"] = 1
    h.config["data_loader"]["batch_size"] = 4
    h.config["general"]["results_dir"] = str(tmp_path)
    h.config["general"]["dev_mode"] = True

    # Only provide `train` and `infer` model_inputs (no `validate` key).
    h.config["model_inputs"] = {
        "train": {
            "data": {
                "dataset_class": "HyraxRandomDataset",
                "data_location": str(tmp_path / "data_train"),
                "primary_id_field": "object_id",
            }
        },
        "infer": {
            "data": {
                "dataset_class": "HyraxRandomDataset",
                "data_location": str(tmp_path / "data_infer"),
                "primary_id_field": "object_id",
            }
        },
    }

    # Configure the underlying random dataset used by tests
    h.config["data_set"]["HyraxRandomDataset"]["size"] = 20
    h.config["data_set"]["HyraxRandomDataset"]["seed"] = 0
    h.config["data_set"]["HyraxRandomDataset"]["shape"] = [2, 3]

    # Percent-based split parameters - these should be applied to the single
    # location `train` dataset and produce a validate split implicitly.
    h.config["data_set"]["train_size"] = 0.6
    h.config["data_set"]["validate_size"] = 0.2
    h.config["data_set"]["test_size"] = 0.2

    # Instead of running full training, validate that the legacy percent-based
    # split path creates both train and validate dataloaders with expected sizes.
    from hyrax.pytorch_ignite import dist_data_loader, setup_dataset

    # Create dataset dict using the same logic as training
    dataset = setup_dataset(h.config)

    assert "train" in dataset

    data_loaders = dist_data_loader(dataset["train"], h.config, ["train", "validate"])

    # Should have created both train and validate loaders
    assert "train" in data_loaders and "validate" in data_loaders

    train_loader, train_indexes = data_loaders["train"]
    validate_loader, validate_indexes = data_loaders["validate"]

    # Assert expected sizes: train 12 (60% of 20), validate 4 (20% of 20)
    assert len(train_indexes) == 12
    assert len(validate_indexes) == 4

    # Finally, run full training to exercise `train.py` end-to-end and ensure
    # the training verb functions correctly with percent-based splits.
    h.train()



from unittest.mock import MagicMock, call, patch

import pytest

from hyrax.pytorch_ignite import setup_dataset


class TestSetupDataset:
    """Tests for the setup_dataset function, specifically focused on iterable dataset handling."""

    def test_setup_dataset_missing_dataset_class_raises_error(self):
        """Test that missing dataset_class raises appropriate RuntimeError."""
        # Create a minimal config that would trigger iterable dataset path
        config = {
            "model_inputs": {
                "train": {
                    "test_dataset": {
                        # Intentionally missing "dataset_class"
                        "data_location": "/some/path"
                    },
                },
                "infer": {
                    "test_dataset": {
                        # Intentionally missing "dataset_class"
                        "data_location": "/some/path"
                    },
                },
            }
        }

        # Mock the functions that would be called before our code
        with patch("hyrax.data_sets.data_provider.generate_data_request_from_config") as mock_generate:
            with patch("hyrax.pytorch_ignite.is_iterable_dataset_requested") as mock_is_iterable:
                # Set up mocks to trigger the iterable dataset path
                mock_generate.return_value = config["model_inputs"]
                mock_is_iterable.return_value = True

                # This should raise RuntimeError with our specific message
                with pytest.raises(RuntimeError) as exc_info:
                    setup_dataset(config)

                assert "dataset_class must be specified in 'model_inputs'." in str(exc_info.value)

    def test_setup_dataset_invalid_dataset_class_raises_error(self):
        """Test that providing an invalid dataset_class raises appropriate RuntimeError."""
        # Create a config with an invalid dataset_class
        config = {
            "model_inputs": {
                "train": {
                    "test_dataset": {
                        "dataset_class": "NonExistentDatasetClass",
                        "data_location": "/some/path",
                    }
                },
                "infer": {
                    "test_dataset": {
                        "dataset_class": "NonExistentDatasetClass",
                        "data_location": "/some/path",
                    },
                },
            }
        }

        # Mock the functions that would be called before our code
        with patch("hyrax.data_sets.data_provider.generate_data_request_from_config") as mock_generate:
            with patch("hyrax.pytorch_ignite.is_iterable_dataset_requested") as mock_is_iterable:
                # Set up mocks to trigger the iterable dataset path
                mock_generate.return_value = config["model_inputs"]
                mock_is_iterable.return_value = True
                # Make the registry lookup fail by simulating the dataset class not being in registry

                # This should raise RuntimeError with our specific message
                with pytest.raises(ValueError) as exc_info:
                    setup_dataset(config)

                assert "Class name NonExistentDatasetClass" in str(exc_info.value)

    def test_setup_dataset_missing_data_location_uses_none(self):
        """Test that missing data_location passes None to dataset constructor."""
        # Create a config with dataset_class but no data_location
        config = {
            "model_inputs": {
                "train": {
                    "test_dataset": {
                        "dataset_class": "HyraxRandomIterableDataset"
                        # Intentionally missing "data_location"
                    },
                },
                "infer": {
                    "test_dataset": {
                        "dataset_class": "HyraxRandomIterableDataset"
                        # Intentionally missing "data_location"
                    },
                },
            }
        }

        # Mock dataset class and registry
        mock_dataset_instance = MagicMock()
        mock_dataset_cls = MagicMock(return_value=mock_dataset_instance)

        with patch("hyrax.data_sets.data_provider.generate_data_request_from_config") as mock_generate:
            with patch("hyrax.pytorch_ignite.is_iterable_dataset_requested") as mock_is_iterable:
                with patch("hyrax.pytorch_ignite.fetch_dataset_class") as mock_fetch_cls:
                    # Set up mocks
                    mock_generate.return_value = config["model_inputs"]
                    mock_is_iterable.return_value = True
                    mock_fetch_cls.return_value = mock_dataset_cls

                    # Call the function
                    result = setup_dataset(config)

                    # Verify the dataset constructor was called with data_location=None
                    expected_call = call(config=config, data_location=None)
                    assert mock_dataset_cls.call_count == 2
                    mock_dataset_cls.assert_has_calls([expected_call, expected_call])
                    assert result["train"] == mock_dataset_instance
                    assert result["infer"] == mock_dataset_instance

    def test_setup_dataset_with_both_keys_present(self):
        """Test normal case where both dataset_class and data_location are present."""
        # Create a complete config
        config = {
            "model_inputs": {
                "train": {
                    "test_dataset": {
                        "dataset_class": "HyraxRandomIterableDataset",
                        "data_location": "/some/valid/path",
                    },
                },
                "infer": {
                    "test_dataset": {
                        "dataset_class": "HyraxRandomIterableDataset",
                        "data_location": "/some/valid/path",
                    },
                },
            }
        }

        # Mock dataset class and registry
        mock_dataset_instance = MagicMock()
        mock_dataset_cls = MagicMock(return_value=mock_dataset_instance)

        with patch("hyrax.data_sets.data_provider.generate_data_request_from_config") as mock_generate:
            with patch("hyrax.pytorch_ignite.is_iterable_dataset_requested") as mock_is_iterable:
                with patch("hyrax.pytorch_ignite.fetch_dataset_class") as mock_fetch_cls:
                    # Set up mocks
                    mock_generate.return_value = config["model_inputs"]
                    mock_is_iterable.return_value = True
                    mock_fetch_cls.return_value = mock_dataset_cls

                    # Call the function
                    result = setup_dataset(config)

                    # Verify the dataset constructor was called with correct parameters
                    expected_call = call(config=config, data_location="/some/valid/path")
                    assert mock_dataset_cls.call_count == 2
                    mock_dataset_cls.assert_has_calls([expected_call, expected_call])
                    assert result["train"] == mock_dataset_instance
                    assert result["infer"] == mock_dataset_instance

    def test_setup_dataset_sets_tensorboardx_logger(self):
        """Test that tensorboardx_logger is properly set on the dataset."""
        # Create a complete config
        config = {
            "model_inputs": {
                "train": {
                    "test_dataset": {
                        "dataset_class": "HyraxRandomIterableDataset",
                        "data_location": "/some/valid/path",
                    },
                },
                "infer": {
                    "test_dataset": {
                        "dataset_class": "HyraxRandomIterableDataset",
                        "data_location": "/some/valid/path",
                    },
                },
            }
        }

        # Mock dataset class, registry, and logger
        mock_dataset_instance = MagicMock()
        mock_dataset_cls = MagicMock(return_value=mock_dataset_instance)
        mock_logger = MagicMock()

        with patch("hyrax.data_sets.data_provider.generate_data_request_from_config") as mock_generate:
            with patch("hyrax.pytorch_ignite.is_iterable_dataset_requested") as mock_is_iterable:
                with patch("hyrax.pytorch_ignite.fetch_dataset_class") as mock_fetch_cls:
                    # Set up mocks
                    mock_generate.return_value = config["model_inputs"]
                    mock_is_iterable.return_value = True
                    mock_fetch_cls.return_value = mock_dataset_cls

                    # Call the function with logger
                    result = setup_dataset(config, tensorboardx_logger=mock_logger)

                    # Verify the logger was set on the dataset
                    assert result["train"].tensorboardx_logger == mock_logger
                    assert result["infer"].tensorboardx_logger == mock_logger



from pathlib import Path

import pytest
from torch import Size

import hyrax


@pytest.fixture(scope="function", params=[".astropy.csv", ".csv", ".fits", ".pq", ".votable"])
def test_hyrax_small_dataset_hscstars(request):
    """Fixture that gives a hyrax object configured to use the small_dataset_hscstars dataset

    Several common table formats are available for this dataset
    """
    this_file_dir = Path(__file__).absolute().parent
    catalog_file = this_file_dir / "test_data" / "small_dataset_hscstars" / f"star_cat_correct{request.param}"

    h = hyrax.Hyrax()
    h.config["model_inputs"] = {
        "train": {
            "data": {
                "dataset_class": "FitsImageDataSet",
                "data_location": str(catalog_file.parent),
                "primary_id_field": "object_id",
            },
        },
        "infer": {
            "data": {
                "dataset_class": "FitsImageDataSet",
                "data_location": str(catalog_file.parent),
                "primary_id_field": "object_id",
            },
        },
    }
    h.config["data_set"]["filter_catalog"] = str(catalog_file)
    h.config["data_set"]["crop_to"] = [20, 20]

    object_id_column_name = "___object_id" if request.param == ".votable" else "# object_id"

    h.config["data_set"]["object_id_column_name"] = object_id_column_name
    h.config["data_set"]["filename_column_name"] = "star_filename"

    return h


def test_prepare(test_hyrax_small_dataset_hscstars):
    """Check that the hsc stars dataset was correctly read, and that basic access
    of FitsImageDataSet returns sane values
    """
    ds = test_hyrax_small_dataset_hscstars.prepare()

    # Dataset has correct length
    for subset in ds:
        a = ds[subset]
        assert len(a) == 10

        # All tensors are the correct size
        for d in a:
            assert d["data"]["image"].shape == Size([1, 20, 20])

        # Selected columns in the original catalog exist
        assert "ira" in a.metadata_fields("data")
        assert "idec" in a.metadata_fields("data")
        assert "SNR" in a.metadata_fields("data")

        # IDs are correct and in the correct order
        assert list(a.ids()) == [
            "36411452835238206",
            "36411452835248579",
            "36411452835249051",
            "36411452835250175",
            "36411457130203411",
            "36411457130204168",
            "36411457130206288",
            "36411457130214646",
            "36411457130215774",
            "36411457130216436",
        ]



from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
import pytest


@pytest.mark.parametrize("shuffle", [True, False])
@pytest.mark.parametrize("split", ["train", "infer", None])
def test_infer_order(loopback_hyrax, split, shuffle):
    """Test that the order of data run through infer
    is correct in the presence of several splits
    """
    h, dataset = loopback_hyrax
    h.config["data_loader"]["shuffle"] = shuffle

    dataset = dataset["infer"]
    inference_results = h.infer()
    inference_result_ids = list(inference_results.ids())
    original_dataset_ids = list(dataset.ids())

    if dataset.is_iterable():
        dataset = list(dataset)
        original_dataset_ids = np.array([str(s["object_id"]) for s in dataset])

    for idx, result_id in enumerate(inference_result_ids):
        dataset_idx = None
        for i, orig_id in enumerate(original_dataset_ids):
            if orig_id == result_id:
                dataset_idx = i
                break
        else:
            raise AssertionError("Failed to find a corresponding ID")

        print(f"orig idx: {dataset_idx}, infer idx: {idx}")
        print(f"orig data: {dataset[dataset_idx]}, infer data: {inference_results[idx]}")
        assert np.all(np.isclose(dataset[dataset_idx]["data"]["image"], inference_results[idx]))


def test_load_model_weights_updates_config_when_auto_detected(tmp_path):
    """Test that config is updated when model_weights_file is auto-detected from train directory"""
    from hyrax.verbs.infer import Infer

    # Create a mock config with no model_weights_file specified
    config = {}
    config["infer"] = {"model_weights_file": None}
    config["train"] = {"weights_filename": "model_weights.pth"}
    config["general"] = {"results_dir": str(tmp_path)}

    # Create a fake train results directory
    train_dir = tmp_path / "20240101-120000-train-abcd"
    train_dir.mkdir(parents=True)
    weights_file = train_dir / "model_weights.pth"
    weights_file.write_text("fake weights content")

    # Create a mock model
    mock_model = MagicMock()

    # Mock find_most_recent_results_dir to return our fake train directory
    with patch("hyrax.config_utils.find_most_recent_results_dir", return_value=train_dir):
        # Call load_model_weights
        Infer.load_model_weights(config, mock_model)

    # Verify that config was updated with the actual weights file path
    assert config["infer"]["model_weights_file"] == str(weights_file)
    # Verify that model.load was called with the correct path
    mock_model.load.assert_called_once_with(weights_file)


def test_load_model_weights_preserves_explicit_config():
    """Test that config is still updated when model_weights_file is explicitly provided"""
    from tempfile import NamedTemporaryFile

    from hyrax.verbs.infer import Infer

    # Create a temporary weights file
    with NamedTemporaryFile(suffix=".pth", delete=False) as tmp_file:
        tmp_file.write(b"fake weights content")
        weights_path = Path(tmp_file.name)

    try:
        # Create a mock config with explicit model_weights_file
        config = {}
        config["infer"] = {"model_weights_file": str(weights_path)}
        config["train"] = {"weights_filename": "model_weights.pth"}

        # Create a mock model
        mock_model = MagicMock()

        # Call load_model_weights
        Infer.load_model_weights(config, mock_model)

        # Verify that config still contains the weights file path (converted to string)
        assert config["infer"]["model_weights_file"] == str(weights_path)
        # Verify that model.load was called with the correct path
        mock_model.load.assert_called_once_with(weights_path)
    finally:
        # Clean up
        weights_path.unlink()



import numpy as np
from astropy.table import Table
from torch.utils.data import Dataset, IterableDataset

from hyrax.data_sets.data_set_registry import HyraxDataset

INVALID_VALUES = {
    "nan": np.nan,
    "inf": np.inf,
    "-inf": -np.inf,
    "none": None,
}
"""Mapping of string representation of invalid values to numpy representations."""


class HyraxRandomDatasetBase:
    """
    This is the base class for the random datasets provided by Hyrax.

    .. warning::

        Direct use of ``HyraxRandomDatasetBase`` is not advised. When working
        with Hyrax, prefer to use ``HyraxRandomDataset`` or ``HyraxRandomIterableDataset``.
    """

    data: np.ndarray
    """The random data samples produced by the dataset."""
    id_list: list
    """A list of sequential numeric IDs for each data sample."""
    provided_labels: list
    """A list of labels randomly selected from the provided list of possible labels."""

    def __init__(self, config, data_location):
        """
        .. py:method:: __init__(config, data_location)

        Initialize the dataset using the parameters defined in the configuration.

        Parameter included for API consistency with other dataset classes, though
        not used by this implementation. All parameters are controlled by the following
        keys under the ``["data_set"]["HyraxRandomDataset"]`` table in the configuration:

        - ``size``: The number of random data samples to produce.
        - ``shape``: The shape of each random data sample as a tuple (e.g. (3, 29, 29) = 3
          layers of 2D data, each layer is 29x29 elements).
        - ``seed``: The random seed to use for reproducibility.
        - ``provided_labels``: A list of possible labels to randomly select from.
          If this is provided, the dataset will randomly select a label for each data sample.
        - ``metadata_fields``: A list of metadata field names. Used to create a metadata
            table with columns corresponding to each field name. All data is numeric.
        - ``number_invalid_values``: The number of invalid values to insert into the data.
        - ``invalid_value_type``: The type of invalid value to insert into the data.
          Valid values are "nan", "inf", "-inf", "none", or a float value.
        """
        # The total number of random data samples produced
        data_size = config["data_set"]["HyraxRandomDataset"]["size"]
        if not isinstance(data_size, int):
            raise ValueError(
                f"Expected integer for `config['data_set']['random_dataset']['size']`, but got {data_size}"
            )

        # The shape of each random data sample as a tuple.
        # i.e. (3, 29, 29) = 3 layers of 2d data, each layer is 29x29 elements.
        data_shape = tuple(config["data_set"]["HyraxRandomDataset"]["shape"])
        if not len(data_shape):
            raise ValueError(
                "Expected `config['data_set']['random_dataset']['data_shape']` to have at least 1 value."
            )

        for e in data_shape:
            if e < 1:
                raise ValueError(
                    f"Expected all values in `config['data_set']['random_dataset']['data_shape']`\
                        to be > 0, but got {data_shape}."
                )
            if not isinstance(e, int):
                raise ValueError(
                    f"Expected all values in `config['data_set']['random_dataset']['data_shape']`\
                        to be integers, but got {data_shape}."
                )

        # Random seed to use for reproducibility
        seed = config["data_set"]["HyraxRandomDataset"]["seed"]
        rng = np.random.default_rng(seed)

        # Note: We raise exceptions if data_size is not an int, so we can assume
        # that turning that into a tuple and adding `data_shape` should work.
        self.data = rng.random((data_size,) + data_shape, np.float32)

        # Start our IDs at a random integer between 0 and 100
        id_start = rng.integers(100)
        self.id_list = np.arange(id_start, id_start + data_size)

        # Randomly insert flawed values (np.nan, np.inf, -np.inf, None, other float)
        num_invalid_values = config["data_set"]["HyraxRandomDataset"]["number_invalid_values"]
        if num_invalid_values:
            # Determine what value to use for invalid values
            invalid_value_type = config["data_set"]["HyraxRandomDataset"]["invalid_value_type"]
            if isinstance(invalid_value_type, str):
                try:
                    invalid_value = INVALID_VALUES[invalid_value_type.lower()]
                except KeyError as err:
                    raise ValueError(
                        f"Invalid value type '{invalid_value_type}' provided. "
                        f"Expected `config['data_set']['random_dataset']['invalid_value_type']` "
                        f"to be one of {list(INVALID_VALUES.keys())}"
                    ) from err
            else:
                if not isinstance(invalid_value_type, float):
                    raise ValueError(
                        f"Expected `config['data_set']['random_dataset']['invalid_value_type']` to be "
                        f"a string or a float, but got {type(invalid_value_type)}."
                    )
                invalid_value = invalid_value_type
            flattened = np.ravel(self.data)
            random_inds = rng.choice(flattened.size, size=40, replace=False)
            flattened[random_inds] = invalid_value

        # If a list of possible labels is provided, create the random label list.
        self.provided_labels = config["data_set"]["HyraxRandomDataset"]["provided_labels"]
        if self.provided_labels:
            self.labels = rng.choice(self.provided_labels, size=data_size)

        meta_fields = config["data_set"]["HyraxRandomDataset"]["metadata_fields"]

        meta = {"object_id": [str(id) for id in self.id_list]}

        for i, field in enumerate(meta_fields):
            meta[field] = np.array(list(range(data_size, 0, -1))) / (i + 2)

        # Create a metadata_table that is used when visualizing data
        metadata_table = Table(meta)

        super().__init__(config, metadata_table)

        self.data_location = data_location

    def get_image(self, idx: int) -> np.ndarray:
        """Get the image at the given index as a NumPy array."""
        return self.data[idx]

    def get_label(self, idx: int) -> str:
        """Get the label at the given index."""
        if self.provided_labels:
            return self.labels[idx]
        return None

    def get_object_id(self, idx: int) -> str:
        """Get the index of the item."""
        return str(self.id_list[idx])


class HyraxRandomDataset(HyraxRandomDatasetBase, HyraxDataset, Dataset):
    """This dataset is stand-in for a map-style dataset.
    It will produce random numpy arrays along with sequential numeric ids and,
    optionally, labels randomly selected from the provided list of possible labels.
    """

    def __getitem__(self, idx: int) -> dict:
        """Get a data sample by index. The returned dictionary will contain the
        following keys:

        - ``index``: The index of the data sample.
        - ``object_id``: The ID of the data sample.
        - ``image``: The data sample as a numpy array.
        - ``label``: The label of the data sample (if provided).

        Parameters
        ----------
        idx : int
            The index of the data sample to retrieve.

        Returns
        -------
        dict
            A dictionary containing the data sample and its metadata.

        """

        ret = {
            "data": {
                "index": idx,
                "object_id": self.get_object_id(idx),
                "image": self.get_image(idx),
            },
            "object_id": self.get_object_id(idx),
        }

        if self.provided_labels:
            ret["data"]["label"] = self.get_label(idx)

        return ret

    def __len__(self):
        """Get the total number of samples in this dataset. This should be return
        the same value as the `size` parameter in the configuration."""
        return len(self.data)

    def ids(self):
        """This function yields IDs for the dataset. It can be used as an iterable
        in a loop, or converted to a list by wrapping the function call in ``list(...)``."""
        for id_item in self.id_list:
            yield str(id_item)


class HyraxRandomIterableDataset(HyraxRandomDatasetBase, HyraxDataset, IterableDataset):
    """This dataset is stand-in for a iterable-style, or streaming, dataset.
    It will produce random numpy arrays and, optionally, labels randomly
    selected from the provided list of possible labels.

    .. note::

        While ids will be generated automatically for this dataset, calling the
        ``ids`` method of this dataset will return the index instead of the id.
    """

    def __iter__(self):
        """Yield the next data sample. The returned dictionary will have the
        following form:

        - ``data``: A dictionary containing the following keys:
        -- ``index``: The index of the data sample.
        -- ``object_id``: The value will be the same as ``index`` for this dataset.
        -- ``image``: The data sample as a numpy array.
        -- ``label``: The label of the data sample (if provided).

        Returns
        -------
        dict
            A dictionary containing a data sample and its metadata.

        """
        for idx, _ in enumerate(self.data):
            ret = {
                "data": {
                    "index": idx,
                    "object_id": self.get_object_id(idx),
                    "image": self.get_image(idx),
                },
                "object_id": self.get_object_id(idx),
            }

            if self.provided_labels:
                ret["data"]["label"] = self.get_label(idx)

            yield ret



import numpy as np
import pytest
import torch
from torch import any, from_numpy, isnan, tensor

import hyrax
from hyrax.data_sets.random.hyrax_random_dataset import HyraxRandomDataset
from hyrax.pytorch_ignite import _handle_nans


class RandomNaNDataset(HyraxRandomDataset):
    """Dataset yielding pairs of random numbers. Requires a seed to emulate
    static data on the filesystem between instantiations"""

    def __init__(self, config, data_location):
        super().__init__(config, data_location)

    def __getitem__(self, idx):
        return from_numpy(self.data[idx])


@pytest.fixture(scope="function", params=["RandomNaNDataset", "HyraxRandomDataset"])
def loopback_hyrax_nan(tmp_path_factory, request):
    """This generates a loopback hyrax instance
    which is configured to use the loopback model
    and a simple dataset yielding random numbers
    """
    results_dir = tmp_path_factory.mktemp("loopback_hyrax_nan")

    h = hyrax.Hyrax()
    h.config["model"]["name"] = "HyraxLoopback"
    h.config["train"]["epochs"] = 1
    h.config["data_loader"]["batch_size"] = 5
    h.config["general"]["results_dir"] = str(results_dir)

    h.config["general"]["dev_mode"] = True
    h.config["model_inputs"] = {
        "train": {
            "data": {
                "dataset_class": request.param,
                "data_location": str(tmp_path_factory.mktemp("data")),
                "primary_id_field": "object_id",
            },
        },
        "infer": {
            "data": {
                "dataset_class": request.param,
                "data_location": str(tmp_path_factory.mktemp("data")),
                "primary_id_field": "object_id",
            },
        },
    }
    h.config["data_set"]["HyraxRandomDataset"]["size"] = 20
    h.config["data_set"]["HyraxRandomDataset"]["seed"] = 0
    h.config["data_set"]["HyraxRandomDataset"]["shape"] = [2, 3]
    h.config["data_set"]["HyraxRandomDataset"]["number_invalid_values"] = 40
    h.config["data_set"]["HyraxRandomDataset"]["invalid_value_type"] = "nan"

    h.config["data_set"]["validate_size"] = 0.2
    h.config["data_set"]["test_size"] = 0.2
    h.config["data_set"]["train_size"] = 0.6

    weights_file = results_dir / "fakeweights"
    with open(weights_file, "a"):
        pass
    h.config["infer"]["model_weights_file"] = str(weights_file)

    dataset = h.prepare()
    return h, dataset["infer"]


def test_nan_handling(loopback_hyrax_nan):
    """
    Test that default nan handling removes nans
    """
    h, dataset = loopback_hyrax_nan
    h.config["data_set"]["nan_mode"] = "quantile"

    inference_results = h.infer()

    if isinstance(dataset[0], dict):
        original_nans = tensor([any(isnan(tensor(item["data"]["image"]))) for item in dataset])
    else:
        original_nans = tensor([any(isnan(item)) for item in dataset])
    assert any(original_nans)

    for result in inference_results:
        assert not any(isnan(result))


def test_nan_handling_zero_values(loopback_hyrax_nan):
    """
    Test that zero nan handling removes nans
    """
    h, dataset = loopback_hyrax_nan
    h.config["data_set"]["nan_mode"] = "zero"

    inference_results = h.infer()

    if isinstance(dataset[0], dict):
        original_nans = tensor([any(isnan(tensor(item["data"]["image"]))) for item in dataset])
    else:
        original_nans = tensor([any(isnan(item)) for item in dataset])
    assert any(original_nans)

    for result in inference_results:
        assert not any(isnan(result))


def test_nan_handling_off(loopback_hyrax_nan):
    """
    Test that when nan handling is off nans appear in output
    """
    h, dataset = loopback_hyrax_nan

    h.config["data_set"]["nan_mode"] = False
    inference_results = h.infer()

    if isinstance(dataset[0], dict):
        original_nans = tensor([any(isnan(tensor(item["data"]["image"]))) for item in dataset])
    else:
        original_nans = tensor([any(isnan(item)) for item in dataset])
    assert any(original_nans)

    result_nans = tensor([any(isnan(item)) for item in inference_results])
    assert any(result_nans)


def test_nan_handling_off_returns_input(loopback_hyrax_nan):
    """Ensure that when nan_mode is False, that the original values passed to
    _handle_nans are returned unchanged."""

    def to_tensor(data_dict):
        data = data_dict.get("data", {})
        if "image" in data and "label" in data:
            image = tensor(data["image"])
            label = data["label"]
            return (image, label)

    h, dataset = loopback_hyrax_nan
    h.config["data_set"]["nan_mode"] = False

    sample_data = dataset[0]

    # If the sample data is a dictionary, convert it to a tuple
    if isinstance(sample_data, dict):
        sample_data = to_tensor(sample_data)

    output = _handle_nans(sample_data, h.config)

    # If the sample was a tuple, check all the elements
    if isinstance(sample_data, tuple):
        assert np.all(np.isclose(output[0], sample_data[0], equal_nan=True))
        assert output[1] == sample_data[1]
    else:
        assert np.all(np.isclose(output, sample_data, equal_nan=True))


def test_nan_handling_tuple_with_three_elements(loopback_hyrax_nan):
    """Test that tuples with more than 2 elements are handled correctly.
    This simulates the AppleCider use case where to_tensor returns (metadata, image, labels)."""
    h, dataset = loopback_hyrax_nan

    # Create a test tuple with 3 elements: metadata (non-tensor),
    # image (tensor with NaNs), labels (non-tensor)
    metadata = {"id": 123, "timestamp": "2024-01-01"}
    image_with_nans = torch.tensor([[1.0, float("nan"), 3.0], [4.0, 5.0, float("nan")]])
    labels = "test_label"

    test_tuple = (metadata, image_with_nans, labels)

    # Test with nan_mode = 'zero'
    h.config["data_set"]["nan_mode"] = "zero"
    output = _handle_nans(test_tuple, h.config)

    # Verify we get back a tuple with 3 elements
    assert isinstance(output, tuple)
    assert len(output) == 3

    # Verify metadata is unchanged
    assert output[0] == metadata

    # Verify image had NaNs replaced with zeros
    assert isinstance(output[1], torch.Tensor)
    assert not torch.any(torch.isnan(output[1]))
    assert torch.all(output[1] == torch.tensor([[1.0, 0.0, 3.0], [4.0, 5.0, 0.0]]))

    # Verify labels are unchanged
    assert output[2] == labels


def test_nan_handling_tuple_multiple_tensors(loopback_hyrax_nan):
    """Test that when a tuple contains multiple tensors, NaN handling is applied
    to all of them."""
    h, dataset = loopback_hyrax_nan

    # Create a test tuple with multiple tensors
    tensor1 = torch.tensor([1.0, float("nan"), 3.0])
    tensor2 = torch.tensor([4.0, 5.0, float("nan")])
    label = "label"

    test_tuple = (tensor1, tensor2, label)

    # Test with nan_mode = 'zero'
    h.config["data_set"]["nan_mode"] = "zero"
    output = _handle_nans(test_tuple, h.config)

    # Verify all tensors had NaN handling applied
    assert len(output) == 3
    assert not torch.any(torch.isnan(output[0]))
    assert not torch.any(torch.isnan(output[1]))
    assert output[2] == label

    # Verify the values are correct
    assert torch.all(output[0] == torch.tensor([1.0, 0.0, 3.0]))
    assert torch.all(output[1] == torch.tensor([4.0, 5.0, 0.0]))


def test_nan_handling_tuple_preserves_order(loopback_hyrax_nan):
    """Test that the order of elements in the tuple is preserved."""
    h, dataset = loopback_hyrax_nan

    # Create a test tuple with mixed types in specific order
    elem1 = "first"
    elem2 = torch.tensor([1.0, float("nan")])
    elem3 = {"key": "value"}
    elem4 = torch.tensor([float("nan"), 2.0])
    elem5 = 42

    test_tuple = (elem1, elem2, elem3, elem4, elem5)

    # Test with nan_mode = 'zero'
    h.config["data_set"]["nan_mode"] = "zero"
    output = _handle_nans(test_tuple, h.config)

    # Verify length and order
    assert len(output) == 5
    assert output[0] == elem1
    assert isinstance(output[1], torch.Tensor)
    assert not torch.any(torch.isnan(output[1]))
    assert output[2] == elem3
    assert isinstance(output[3], torch.Tensor)
    assert not torch.any(torch.isnan(output[3]))
    assert output[4] == elem5



import logging

import pytest

from hyrax import plugin_utils
from hyrax.models import hyrax_model
from hyrax.models.model_registry import fetch_model_class


def test_import_module_from_string():
    """Test the import_module_from_string function."""
    module_path = "builtins.BaseException"

    model_cls = plugin_utils.import_module_from_string(module_path)

    assert model_cls.__name__ == "BaseException"


def test_import_module_from_string_no_base_module():
    """Test that the import_module_from_string function raises an error when
    the base module is not found."""

    module_path = "nonexistent.BaseException"

    with pytest.raises(ModuleNotFoundError) as excinfo:
        plugin_utils.import_module_from_string(module_path)

    assert "Module nonexistent not found" in str(excinfo.value)


def test_import_module_from_string_no_submodule():
    """Test that the import_module_from_string function raises an error when
    a submodule is not found."""

    module_path = "builtins.nonexistent.BaseException"

    with pytest.raises(ModuleNotFoundError) as excinfo:
        plugin_utils.import_module_from_string(module_path)

    assert "Module builtins.nonexistent not found" in str(excinfo.value)


def test_import_module_from_string_no_class():
    """Test that the import_module_from_string function raises an error when
    a class is not found."""

    module_path = "builtins.Nonexistent"

    with pytest.raises(AttributeError) as excinfo:
        plugin_utils.import_module_from_string(module_path)

    assert "Unable to find Nonexistent in module" in str(excinfo.value)


def test_fetch_model_class():
    """Test the fetch_model_class function."""
    config = {"model": {"name": "builtins.BaseException"}}

    model_cls = fetch_model_class(config)

    assert model_cls.__name__ == "BaseException"


def test_fetch_model_class_no_model():
    """Test that the fetch_model_class function raises an error when no model
    is specified in the configuration."""

    config = {"model": {"name": ""}}

    with pytest.raises(RuntimeError) as excinfo:
        fetch_model_class(config)

    assert "A model class name or path must be provided" in str(excinfo.value)


def test_fetch_model_class_false_model():
    """Test that the fetch_model_class function raises an error when model
    is set to false in the configuration."""

    config = {"model": {"name": ""}}

    with pytest.raises(RuntimeError) as excinfo:
        fetch_model_class(config)

    assert "A model class name or path must be provided" in str(excinfo.value)


def test_fetch_model_class_no_model_cls():
    """Test that an exception is raised when a non-existent model class is requested."""

    config = {"model": {"name": "builtins.Nonexistent"}}

    with pytest.raises(AttributeError) as excinfo:
        fetch_model_class(config)

    assert "Unable to find Nonexistent in module" in str(excinfo.value)


def test_fetch_model_class_not_in_registry():
    """Test that an exception is raised when a model is requested that is not in the registry."""

    config = {"model": {"name": "Nonexistent"}}

    with pytest.raises(ValueError) as excinfo:
        fetch_model_class(config)

    assert "not found in registry and is not a full import path" in str(excinfo.value)


def test_fetch_model_class_false_logs_registered_models(caplog):
    """Test that the fetch_model_class function logs registered models when
    model is set to false."""

    config = {"model": {"name": ""}}

    with caplog.at_level(logging.ERROR):
        with pytest.raises(RuntimeError):
            fetch_model_class(config)

    # Check that the error message contains expected information
    assert "No model name was provided" in caplog.text
    assert "h.set_config('model.name'" in caplog.text
    assert "Currently registered models:" in caplog.text


def test_fetch_model_class_in_registry():
    """Test that a model class is returned when it is in the registry."""

    # make a no-op model that will be added to the model registry
    @hyrax_model
    class NewClass:
        pass

    config = {"model": {"name": "NewClass"}}
    model_cls = fetch_model_class(config)

    assert model_cls.__name__ == "NewClass"
